about = {
    "__title__": "naz",
    "__description__": "Naz is an SMPP client.",
    "__url__": "https://github.com/komuw/naz",
    "__version__": "0.0.0.6",
    "__author__": "komuW",
    "__author_email__": "komuw05@gmail.com",
    "__license__": "MIT",
}
